require 'open-uri'
require 'flux/git'
require 'github_api'

module Flux
  module CLI
    class Review < Thor
      G = Flux::Git
      B = G::Branch

      PULL_REQUEST_HEADERS = %w(>NUM >STORY AUTHOR DATE TITLE)

      include Flux::Util

      namespace 'review'

      desc "diff [BRANCH]", "show interdiff between last 2 reviews"
      method_option :from,  :type => :numeric, :aliases => '-f'
      method_option :to,    :type => :numeric, :aliases => '-t'
      method_option :color, :type => :boolean, :aliases => '-c'
      def diff(branch_name = B.current.name)
        reqs = Github::PullRequest.all(:branch_name => branch_name)

        if options[:from] && options[:to]
          to   = reqs.find { |r| r.number.to_i == options[:to].to_i }
          from = reqs.find { |r| r.number.to_i == options[:from].to_i }
        else
          to, from = reqs.first(2)
        end

        if from
          to_diff = Tempfile.new("to_diff")
          to_diff.write(to.diff)
          to_diff.flush

          from_diff = Tempfile.new("from_diff")
          from_diff.write(from.diff)
          from_diff.flush

          cmd =  "interdiff #{from_diff.path} #{to_diff.path}"
          cmd << "| colordiff" if options[:color]

          system cmd
        end
      end

      desc "request [BRANCH]", "request a review"
      def request(branch = B.current)
        branch = G::Branch(branch)

        invoke 'git:fetch'
        invoke 'branch:checks:synced', [branch]
        invoke 'branch:checks:rebased', [branch], :parent => 'origin/master'

        story  = Flux::PT::Project.
          current.stories.find(branch.config('story_id'))
        old    = Github::PullRequest.all(:branch_name => branch.name)
        last   = old.first

        last.close if last && last.open?

        body = {:branch_name => branch.name,
                :story_url   => story.url,
                :iterations  => old.map(&:html_url)}
        data = {:base  => 'master',
                :head  => branch.top.sha,
                :title => story.name,
                :body  => body}

        Github::PullRequest.create data
      end

      desc "list", "list pull requests for the current project"
      def list
        list_pull_requests Github::PullRequest.open
      end

      desc "history [BRANCH]", "show review history for branch"
      def history(branch_id = B.current.name)
        reqs = Github::PullRequest.all(:branch_name => branch_id)

        list_pull_requests reqs
      end

      private

      def config
        Flux.environment
      end

      def list_pull_requests(pull_requests)
        table =
          [PULL_REQUEST_HEADERS] +
          pull_requests.map { |p|
            [p.number, Flux::PT::Story.id_from_url(p.body[:story_url]), p.author,
             Time.parse(p.created_at).strftime("%F %R"), p.title]
          }

        puts_table table
      end

      module Github
        def self.config
          Flux.environment['github']
        end

        def self.gh
          ::Github.new(:login    => Github.config['username'],
                       :password => Github.config['password'])
        end

        class PullRequest
          class << self
            def all(opts = {})
              pull_requests(opts.merge(:state => 'open')) +
                pull_requests(opts.merge(:state => 'closed'))
            end

            def create(data)
              pr.create_request Github.config['repo_user'],
                                Github.config['repo_name'],
                                PullRequest.new(data).to_hash
            end

            def open(opts = {})
              pull_requests({:state => 'open'}.merge(opts))
            end

            def pr
              Github.gh.pull_requests
            end

            def update(id, data)
              pr.update_request Github.config['repo_user'],
                                Github.config['repo_name'],
                                id,
                                data.dup
            end

            private

            def pull_requests(opts = {})
              branch_name = opts.delete(:branch_name)

              reqs = pr.pull_requests(Github.config['repo_user'],
                                      Github.config['repo_name'],
                                      opts).map { |r|
                r[:author]   = r[:user][:login]
                r[:body]     = ReviewBody.from_markdown(r[:body])

                PullRequest.new(r)
              }

              if branch_name
                reqs.select { |r| r.body[:branch_name] == branch_name }
              else
                reqs
              end
            end
          end

          def initialize(data)
            @data = data
          end

          def close
            @data[:state] = 'closed'

            self.class.update number, {:state => @data[:state]}
          end

          def diff
            auth = [Github.config['username'], Github.config['password']]

            open(@data[:diff_url], :http_basic_authentication => auth).read
          end

          def method_missing(name, *args)
            if @data.has_key?(name.to_sym)
              @data[name.to_sym]
            else
              super
            end
          end

          def open?
            state == 'open'
          end

          def to_hash
            @data.merge(:body => ReviewBody.to_markdown(@data[:body]))
          end
        end
      end

      module ReviewBody
        def self.from_markdown(md)
          if (md || '').include?(signoff)
            _, branch_name, story_url, iter = md.
              gsub(signoff, '').
              split(/### [^\n]+/).map(&:strip)

            if iter
              _, iterations = iter.split('*').map(&:strip)

              iterations = Array(iterations)
            else
              iterations = []
            end
          else
            branch_name = nil
            story_url   = "http://mojotech.com"
            iterations  = []
          end

          {:branch_name => branch_name,
           :story_url   => story_url,
           :iterations  => iterations}
        end

        def self.to_markdown(data)
          tmpl = <<MARKDOWN
### Branch

#{data[:branch_name]}

### Story

#{data[:story_url]}

MARKDOWN

          if data[:iterations].empty?
            tmpl << signoff << "\n"
          else
            tmpl << <<ITERATIONS
### Prior Versions

#{"* " + data[:iterations].join("\n* ") + "\n"}

#{signoff}
ITERATIONS
          end
        end

        def self.signoff
          "*Your friendly neighborhood Flux*"
        end
      end
    end
  end
end
