require 'grit'

module Flux
  module Git
    def self.git(path = repo_path)
      Grit::Git.new(path)
    end

    def self.repo(path = repo_path)
      Grit::Repo.new(path)
    end

    def self.repo_path(start_dir = Dir.pwd)
      Flux.find_upwards('.git', start_dir) or
        raise RCSError, "Couldn't find git repo starting at #{start_dir}."
    end

    def self.Branch(branch, repo_path = Git.repo_path)
      case branch
      when Branch
        branch
      when String
        remotes = Git.repo(repo_path).remote_list

        if remotes.find { |r| branch =~ %r{^#{r}/} }
          args = branch.split('/', 2).reverse << repo_path

          Branch.remote(*args)
        else
          Branch.local(branch, repo_path)
        end
      else
        raise "Unable to convert object to branch: #{branch.inspect}"
      end
    end

    class Branch < Struct.new(:repo_path, :remote, :name)
      def self.current(repo_path = Git.repo_path)
        Branch.local(Git.repo(repo_path).head.name, repo_path)
      end

      def self.remote(name, remote = 'origin', repo_path = Git.repo_path)
        new(repo_path, remote, name)
      end

      def self.local(name, repo_path = Git.repo_path)
        new(repo_path, nil, name)
      end

      def initialize(repo_path, remote, name)
        super

        @git  = Git.git(repo_path)
        @repo = Git.repo(repo_path)
      end

      def config(var, value = nil)
        gitvar = "branch.#{name}.#{var.gsub('_', '')}"

        if value
          @git.config({:add => true}, gitvar, value)
        else
          @git.config({}, gitvar).chomp
        end
      end

      def create(parent)
        tap { @git.branch({}, @name, parent.name) }
      end

      def checkout
        tap { @git.checkout({}, @name) }
      end

      def fqdn
        remote ? "#{remote}/#{name}" : name
      end

      def include?(branch)
        @repo.commits_between(fqdn, branch.fqdn).empty?
      end

      alias_method :includes?, :include?

      def top
        @repo.commits(fqdn, 1).first
      end

      def publish(upstream)
        tap {
          @git.push({}, upstream.remote, "+#{@name}:#{upstream.name}")
        }
      end

      def track(upstream)
        tap { @git.branch({:set_upstream => true}, @name, upstream.fqdn) }
      end

      def synced?
        upstream.top.sha == top.sha
      end

      def upstream
        remote = config('remote')
        name   = config('merge').sub('refs/heads/', '')

        Branch.remote(name, remote, repo_path)
      end
    end
  end
end
