module Flux
  module PT
    def self.config
      Flux.environment['pivotal_tracker']
    end

    class Project
      attr_reader :id

      def self.current
        Project.new(PT.config['project_id'])
      end

      def initialize(id)
        @id = id
      end

      def members
        ::PivotalTracker::Membership.all(self).tap { |members|
          class << members
            def me
              find { |m| m.email == PT.config['email'] }
            end
          end
        }
      end

      def stories
        Story::Finders.new(self)
      end
    end

    class Story
      STATES_W_ESTIMATE = %w(started finished delivered accepted rejected)
      STATES            = %w(unscheduled unstarted) + STATES_W_ESTIMATE

      extend Forwardable

      class Finders
        def initialize(project)
          @project = project
        end

        def find(id)
          story = ::PivotalTracker::Story.find(id, @project.id) or
            raise TrackerError, "Couldn't find story #{id}."

          Story.new(story)
        end

        def scheduled
          ::PivotalTracker::Iteration.
            current_backlog(@project).
            map(&:stories).
            flatten.map { |s| Story.new(s) }
        end

        def owned_by(member)
          ::PivotalTracker::Story.all(@project, :mywork => member.name).
            map { |s| Story.new(s) }
        end

        def unassigned
          scheduled.reject(&:owned_by)
        end
      end

      def_delegators :@story, :id, :name, :owned_by, :current_state, :url

      alias_method :state, :current_state

      def self.id_from_url(url)
        URI.parse(url).path.split('/').last
      end

      def initialize(story)
        @story = story
      end

      def estimate
        (@story.estimate && @story.estimate < 0) ? nil : @story.estimate
      end

      alias_method :est, :estimate

      def update(attrs)
        if attrs[:current_state]
          unless STATES.include?(attrs[:current_state])
            raise TrackerError, "Invalid state: #{attrs[:current_state]}"
          end

          if STATES_W_ESTIMATE.include?(attrs[:current_state]) &&
             ! estimate &&
             ! attrs[:estimate]
            raise TrackerError,
                  "Need an estimate for state `#{attrs[:current_state]}'."
          end
        end

        @story.update attrs
      end
    end
  end
end
