module Flux
  module Util
    # Initialize a new Table.
    #
    # @param [Array<Array>] raw_data the table's annotated data
    # @param field_sep the string used to separate columns
    # @param [Integer] max_width the maximum width for this table.
    #   Lines exceeding this width will be truncated.
    # @param shell the shell object used by the current task
    def puts_table(raw_data, field_sep = '  ', max_width = nil)
      $stdout.puts Table.new(shell, raw_data, field_sep, max_width)
    end

    # A table pretty printer that allows us to take a peek at the data after
    # it's been stripped of formatting information. It exists to make testing
    # easier.
    #
    # If any cell in the first row is prefixed with '>', then that column's
    # contents will be right-aligned. This is useful when showing numbers.
    class Table
      ALIGN_RIGHT = '>'

      attr_reader :field_sep, :raw_data, :shell, :max_width

      # @see Flux::Util#puts_table
      def initialize(shell, raw_data, field_sep, max_width)
        @shell     = shell
        @raw_data  = raw_data
        @field_sep = field_sep
        @max_width = max_width || terminal_width
      end

      # @return whether a given column's contents will be aligned to the right
      def align_right?(index)
        meta[index].include?(ALIGN_RIGHT)
      end

      # @return all rows but the first
      def body
        @body ||= raw_data[1..-1]
      end

      # @return the table data, stripped of format information
      def data
        @data ||= body.unshift(headers)
      end

      # @return the table's first row
      def headers
        @headers = raw_data.first.each_with_index.map { |e, i|
          align_right?(i) ? e[1..-1] : e
        }
      end

      # @return the table's format information
      def meta
        @meta ||= @raw_data.first.inject([]) { |a, e|
          a << (e =~ /^>/ ? ALIGN_RIGHT : '')
        }
      end

      def to_s
        @s ||=
          begin
            widths  = data.transpose.inject([]) { |a, col|
              a << col.max { |a, b| a.to_s.size <=> b.to_s.size }.to_s.size
            }

            # we don't need the last column's width if it is to be left-aligned
            widths[-1] = nil unless align_right?(-1)

            format  = meta.each_with_index.inject('') { |a, (e, i)|
              w     = widths[i] ? widths[i].to_s : ''
              align = align_right?(i) ? '' : '-'

              a << "#{field_sep}%" << align << w << 's'
            }.strip

            max_width

            data.inject('') { |a, r|
              l =
              a << truncate(format % r, max_width) << "\n"
            }
          end
      end

      private

      def truncate(string, width)
        shell.send(:truncate, string, width)
      end

      def terminal_width
        shell.send(:terminal_width)
      end
    end
  end
end
